import os

import click
import colorama as co
from tomlkit import dumps

from yatta.completion_helpers import get_figlet_fonts, get_table_formats
from yatta.config import Config
from yatta.utils import get_app_dirs

DATA_DIR, CONFIG_DIR, CACHE_DIR = get_app_dirs()


@click.group()
def config():
    """
    Configure yatta.
    """


@config.command()
@click.option(
    "-s", "--shell", required=True, type=click.Choice(["bash", "zsh", "fish"])
)
def completions(shell):
    """
    Generate shell completes for yatta.
    """
    home = os.environ.get("HOME")
    comp_file = os.path.join(CACHE_DIR, "_completions")

    if shell == "bash":
        shell_config = os.path.join(home, ".bashrc")
        os.system(f"_YATTA_COMPLETE=source_bash yatta > {comp_file}")
        print(
            "\nThe tab completion script has been generated! To enable, "
            + f"add the following to {shell_config} and open a new shell:\n"
        )
        print(f"# source yatta completions\nsource {comp_file}")
    elif shell == "zsh":
        shell_config = os.path.join(home, ".zshrc")
        os.system(f"_YATTA_COMPLETE=source_zsh yatta > {comp_file}")
        print(
            "\nThe tab completion script has been generated! To enable, "
            + f"add the following to {shell_config} and open a new shell:\n"
        )
        print(f"# source yatta completions\nsource {comp_file}")
    else:
        os.system(f"_YATTA_COMPLETE=source_zsh yatta > {comp_file}")
        comp_dir = os.path.join(home, ".config/fish/completions/")
        try:
            assert os.path.exists(comp_dir)
            os.system(f"cp {comp_file} {comp_dir}")
            print(
                "The tab completion script has been generated! "
                + "Open a new shell to enable completions."
            )
        except AssertionError:
            print(
                "\nYour fish completion direction is missing, please create it here:\n"
                + f"\n{comp_dir}"
            )


@config.command()
@click.argument("font", type=click.STRING, autocompletion=get_figlet_fonts)
def font(font):
    """
    Select figlet font for displaying stopwatch.
    """
    Config().set_user_value("formatting", "figlet_font", font)


@config.command()
@click.argument("style", type=click.STRING, autocompletion=get_table_formats)
def table_style(style):
    """
    Select tabulate table style.
    """
    Config().set_user_value("formatting", "table_style", style)


@config.command()
@click.argument("columns", type=click.INT)
def plot_columns(columns):
    """
    Set maximum number of columns for plots.
    """
    Config().set_user_value("plotting", "columns", columns)


@config.command()
@click.argument("run_in_background", type=click.BOOL)
def background(run_in_background):
    """
    Always run yatta in the background.
    """
    Config().set_user_value("general", "run_in_background", run_in_background)


@config.command()
@click.argument("show", type=click.BOOL)
def show_legend(show):
    """
    Always display a legend with applicable plots.
    """
    Config().set_user_value("plotting", "show_legend", show_legend)


@config.command()
@click.option("-d", "--defaults", is_flag=True, help="Print the default settings.")
def list(defaults):
    """
    List current user settings.
    """
    co.init(autoreset=True)
    if defaults:
        print(f"\n{co.Fore.BLUE}{dumps(Config().default)}")
    else:
        print(f"\n{co.Fore.GREEN}{Config()}")


@config.command()
@click.confirmation_option(
    prompt="Are you sure you want to reset the application settings?"
)
def restore_defaults():
    """
    Restore settings to default values.
    """
    Config().restore_defaults()



import logging
from datetime import datetime

import click
import parsedatetime as pdt
from sqlalchemy.exc import IntegrityError

from yatta import db as db
from yatta.completion_helpers import get_matching_tasks

logger = logging.getLogger(__name__)


@click.group()
def edit():
    """
    Edit attributes of tasks and records.
    """
    pass


@edit.command()
@click.argument("task_name_or_id", type=click.STRING, autocompletion=get_matching_tasks)
@click.option("-n", "--name", help="Change the name of the task.")
@click.option("-t", "--tags", default="", help="Add relevant tags to task.")
@click.option("-d", "--description", default="", help="Add additional task info.")
def tasks(task_name_or_id, name=None, tags=None, description=None):
    """
    Edit task attributes.
    """
    integrity_error_msg = (
        "Edits were not saved! It's possible you tried "
        + "to change the task name to that of an existing "
        + "task."
    )

    query = db.get_tasks(task_name_or_id)
    _task = query.first()
    if not _task:
        print(f"Task '{task_name_or_id}' does not exist.")
        return
    if name or tags or description:
        if name:
            try:
                _task.name = name
                db.session.commit()
            except IntegrityError:
                db.session.rollback()
                logger.warning(integrity_error_msg)
                return
        if tags:
            _task.tags = tags
            # TODO: is there a better way to handle this?
            db.session.commit()
        if description:
            _task.description = description
            db.session.commit()
        print(_task)
    else:
        MARKER = (
            "\n\n-----\nFORMAT: NAME,TAGS,DESCRIPTION ("
            + "Separate tags with '-')\n"
            + "EXAMPLE: fix bicycle,personal-hobbies-fitness,"
            + "fix the flat from my last ride\n-----"
        )
        task_str = f"{_task.name},{_task.tags},{_task.description}{MARKER}"
        updates = click.edit(task_str)
        if updates:
            try:
                _task.name, _task.tags, _task.description = (
                    updates.split(MARKER)[0].strip().split(",")
                )
                db.session.commit()
                print(_task)
            except IntegrityError:
                logger.warning(integrity_error_msg)
        else:
            print("No updates were specified.")


# TODO: Catch common errors from parsedatetime
@edit.command()
@click.argument("record_id")
@click.option("-s", "--start-time", help="Time when task was started.")
@click.option("-e", "--end-time", help="Time when the task was stopped.")
def records(record_id, start_time=None, end_time=None):
    """
    Edit record details.
    """
    if record_id == "0":
        query = db.get_records()
        record_id = query.all()[-1].id
    query = db.get_records(record_id=record_id)
    _record = query.first()
    if not _record:
        print(f"Record with ID '{record_id}' does not exist.")

    cal = pdt.Calendar()
    if start_time:
        start = datetime(*cal.parse(start_time)[0][:6])
        end = _record.end
    if end_time:
        end = datetime(*cal.parse(end_time)[0][:6])
        start = _record.start
    if not start_time and not end_time:
        MARKER = (
            "\n\n-----\nFORMAT: START,END\n"
            + "EXAMPLES:\n\t2020-03-21 13:01:00,2020-03-21 14:01:00\n"
            + "\tToday at noon,now\n"
            + "START and END must be interpretable by: "
            + "https://github.com/bear/parsedatetime\n-----"
        )
        record_str = f"{_record.start},{_record.end}{MARKER}"
        updates = click.edit(record_str)
        if updates:
            start, end = updates.split(MARKER)[0].strip().split(",")
            start = datetime(*cal.parse(start)[0][:6])
            end = datetime(*cal.parse(end)[0][:6])
        else:
            print("No updates were specified.")
            return

    duration = (end - start).seconds
    if end < start:
        logger.error("End time is before start time. No changes have been made.")
        return
    _record.start = start
    _record.end = end
    _record.duration = duration
    db.session.commit()
    print(_record)



import atexit
import curses
import logging
import os
import signal
import sys
import time
from datetime import datetime

import colorama as co

from yatta import db as db
from yatta.utils import (
    _read_tmp_info,
    get_app_dirs,
    stopwatch,
    time_figlet_print,
    time_print,
)

DATA_DIR, CONFIG_DIR, CACHE_DIR = get_app_dirs()
TMP_FILE = os.path.join(CACHE_DIR, "active_task")
PID_FILE = os.path.join(CACHE_DIR, "yatta.pid")

logger = logging.getLogger(__name__)


class Daemon:
    """
    A generic daemon class.
    Credit: https://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/

    Usage: subclass the daemon class and override the run() method.
    """

    def __init__(self, pidfile):
        self.pidfile = pidfile

    def daemonize(self):
        """Deamonize class. UNIX double fork mechanism."""

        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write("fork #1 failed: {0}\n".format(err))
            sys.exit(1)

        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:

                # exit from second parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write("fork #2 failed: {0}\n".format(err))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, "r")
        so = open(os.devnull, "a+")
        se = open(os.devnull, "a+")

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)

        pid = str(os.getpid())
        with open(self.pidfile, "w+") as f:
            f.write(pid + "\n")

    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """Start the daemon."""

        # Check for a pidfile to see if the daemon already runs
        try:
            with open(self.pidfile, "r") as pf:

                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if pid:
            message = "pidfile {0} already exist. " + "Daemon already running?\n"
            sys.stderr.write(message.format(self.pidfile))
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """Stop the daemon."""

        # Get the pid from the pidfile
        try:
            with open(self.pidfile, "r") as pf:
                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if not pid:
            message = "pidfile {0} does not exist. " + "Daemon not running?\n"
            sys.stderr.write(message.format(self.pidfile))
            return  # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            e = str(err.args)
            if e.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print(str(err.args))
                sys.exit(1)

    def restart(self):
        """Restart the daemon."""
        self.stop()
        self.start()

    def run(self):
        """You should override this method when you subclass Daemon.

        It will be called after the process has been daemonized by
        start() or restart()."""


class StopwatchDaemon(Daemon):
    def __init__(self, pidfile, taskname=None):
        Daemon.__init__(self, pidfile)
        self.taskname = taskname

    def run(self):
        stopwatch(self.taskname)


def daemon_start(taskname):
    daemon = StopwatchDaemon(PID_FILE, taskname)
    # start() will execute the hijacked run() func
    daemon.start()


def daemon_stop():
    daemon = StopwatchDaemon(PID_FILE)
    daemon.stop()
    try:
        start, end, duration, taskname = _read_tmp_info()
    except FileNotFoundError:
        print("No tasks are being tracked right now.")
        return
    # create record
    record = db.Record(task_name=taskname, start=start, end=end, duration=duration)
    # at this point, the task has already been added to db in track(),
    # just need to fetch it
    task = db.get_tasks(taskname).first()
    db.add_record(task, record)
    db.update_task_total(task)
    co.init(autoreset=True)
    print(
        f"\n{co.Fore.GREEN}Worked on {task.name} for {record.duration/3600:.2f}"
        + f"hrs ({time_print(record.duration)}) \u2714"
    )
    os.remove(TMP_FILE)


# FIXME: #10 manually killing yatta with "kill" command doesn't remove pid file
def daemon_status():
    co.init(autoreset=True)
    if os.path.exists(PID_FILE):
        start, end, duration, taskname = _read_tmp_info()
        print(
            f"\n\t{co.Fore.BLUE}Active task: {co.Fore.GREEN}{taskname}"
            + f"\n\t{co.Fore.BLUE}Start: {co.Fore.GREEN}{start}"
            + f"\n\t{co.Fore.BLUE}Current: {co.Fore.GREEN}{end}"
            + f"\n\t{co.Fore.BLUE}Duration: {co.Fore.GREEN}{time_print(int(duration))}"
        )
    else:
        print("No tasks are being tracked right now.")


def dummy_stopwatch(taskname, font):
    """
    Start a (mostly) daemon-naive stopwatch to display w/ curses.
    i.e., it does not check the active_task file
    """

    def show_time(stdscr, taskname, font):
        QUIT_KEY = ord("q")
        curses.echo()
        curses.use_default_colors()
        curses.curs_set(0)
        stdscr.timeout(0)
        try:
            stdscr.addstr(font.renderText(taskname))
        except curses.error as e:
            stdscr.clear()
            stdscr.addstr(font.renderText(taskname[:11] + "..."))
            logger.warning(e)
        start_time = datetime.now()
        while os.path.exists(PID_FILE):
            count = int(time.time() - start_time.timestamp())
            stdscr.insstr(time_figlet_print(font, count))
            ch = stdscr.getch()
            stdscr.refresh()
            # FIXME: #11 task summary is printed awkwardly over curses
            # text on task completion
            if ch == QUIT_KEY:
                daemon_stop()
            time.sleep(1)

    curses.wrapper(show_time, taskname, font)



import logging
from datetime import datetime, timedelta

import colorama as co
import numpy as np
import pandas as pd

from yatta import utils as utils

logger = logging.getLogger(__name__)

plot_unit = "▇"
co.init(autoreset=True)

days = {0: "mon", 1: "tue", 2: "wed", 3: "thu", 4: "fri", 5: "sat", 6: "sun"}

fcolors = [
    co.Fore.LIGHTGREEN_EX,
    co.Fore.LIGHTBLUE_EX,
    co.Fore.LIGHTMAGENTA_EX,
    co.Fore.LIGHTYELLOW_EX,
    co.Fore.LIGHTRED_EX,
    co.Fore.LIGHTCYAN_EX,
    co.Fore.LIGHTBLACK_EX,
    co.Fore.GREEN,
    co.Fore.BLUE,
    co.Fore.MAGENTA,
    co.Fore.YELLOW,
    co.Fore.RED,
    co.Fore.CYAN,
]


def _weekday_to_label(weekday):
    """
    Convert weekday index to string weekday

    Args:
        weekday (int): Day of the week [0-6]

    Returns:
        (string): Weekday as "mon", "tue", etc.
    """
    return days[weekday]


def _month_split(year, month):
    """
    Split month into calendar weeks.

    Args:
        year (int)
        month (int)

    Returns:
        weeks (dict): Dict of tuples with start/end datetime for each calendar week.
    """
    _date = datetime(year, month, 1)
    dow = _date.weekday()
    weeks = {}
    i = 0
    while _date.month == month:
        year, month, day = _date.timetuple()[:3]
        try:
            weeks[i] = (
                datetime(year, month, day),
                datetime(year, month, day + (6 - dow)),
            )
        except ValueError:
            # adjust end datetime if it would end in a new month
            weeks[i] = (
                datetime(year, month, day),
                datetime(year, month + 1, 1) - timedelta(days=1),
            )
        try:
            _date += timedelta(days=7)
        except ValueError:
            # adjust end datetime if it would end in a new month
            _date = datetime(year, month + 1, 1) - timedelta(days=1)
        i += 1

    return weeks


def _preproc_data(data, period, start_date):
    """
    Take dataframe returned from yatta.db.query_to_df() and format for
    timesheet or plotting in hbar_stack()

    Args:
        data (pd.DataFrame): Dataframe from yatta.db.query_to_df()

    Returns:
        data (pd.DataFrame): Dataframe with string index labels and unique tasks as
                             columns
    """
    if period == "day":
        df = data[
            (data["start"] >= start_date)
            & (data["start"] < start_date + timedelta(days=1))
        ]
        if df.empty:
            return df
        df = df.groupby(["task_name", "task_id"]).sum().droplevel("task_id").T
        df.columns = df.columns.values
        data_fmt = df
    elif period == "week":
        df = data[
            (data["start"] >= start_date)
            & (data["start"] < start_date + timedelta(days=7))
        ]
        if df.empty:
            return df
        # break up into days, format df
        group = df.groupby(["start", "task_name"])
        df = group.sum().reset_index()
        df["weekday"] = df["start"].dt.weekday.apply(_weekday_to_label)
        # build properly formatted dataframe
        data_fmt = pd.DataFrame(index=days.values(), columns=df["task_name"].unique())
        for col in data_fmt.columns:
            task_df = df.loc[df["task_name"] == col]
            data_fmt.loc[task_df["weekday"], col] = task_df.groupby(["weekday"]).sum()[
                "duration"
            ]
        data_fmt = data_fmt.dropna(axis=0, how="all").fillna(0)
    elif period == "month":
        year, month, day = start_date.timetuple()[:3]
        df = data[
            (data["start"] >= start_date)
            & (data["start"] < datetime(year, month + 1, day))
        ].sort_values(["start"])
        if df.empty:
            return df
        # Split data into calendar weeks...
        df["week"] = 0
        weeks = _month_split(year, month)
        for week, week_range in weeks.items():
            df.loc[
                (df["start"] >= week_range[0]) & (df["start"] < week_range[1]), "week"
            ] = week
        df = df.groupby(["week", "task_name"]).sum()
        tmp = pd.DataFrame(
            columns=df.index.get_level_values(1).unique().values,
            index=df.index.get_level_values(0).unique().values,
        )
        data_fmt = df.loc[df.index.get_level_values(0) == 0].T
        for col in tmp.columns:
            for week in tmp.index:
                val = df.loc[
                    (df.index.get_level_values(1) == col)
                    & (df.index.get_level_values(0) == week),
                    "duration",
                ].values
                try:
                    tmp.loc[week, col] = val[0]
                except IndexError:
                    tmp.loc[week, col] = 0
        # convert from week index to date range
        if year == datetime.now().year:
            tmp.index = [
                f"{weeks[i][0].strftime('%d')}-{weeks[i][1].strftime('%d %b')}"
                for i in tmp.index
            ]
        else:
            tmp.index = [
                f"{weeks[i][0].strftime('%d')}-{weeks[i][1].strftime('%d %b %Y')}"
                for i in tmp.index
            ]

        data_fmt = tmp
    else:
        logger.error("Invalid plot period!")

    return data_fmt


def _normalize_data(y, columns=50, norm_factor=1):
    """
    Normalize dependent variable to screen width.

    Args:
        y (np.array): Numpy array of ints or floats.
        columns (int): Maximum screen columns.

    Returns:
        y (np.array): Normalized list of ints.
    """
    y = (y / norm_factor) * columns
    y = [int(np.round(val, 0)) for val in y]
    return y


def hbar(data, columns=50):
    """
    Print horizontal bar chart to stdout.

    Args:
        data (pd.Series): Data to plot.
        columns (int): Maximum screen columns.

    Returns:
        None
    """
    norm_factor = data.values.max()
    data_norm = data.apply(_normalize_data, args=([columns, norm_factor]))
    data_norm = data_norm.T.sort_values("duration", ascending=False).T
    print()
    for n, col in enumerate(data_norm.columns):
        val = data_norm[col].values[0]
        s = plot_unit * val
        print(
            f"{co.Fore.RESET}{utils.time_print(data[col].values[0])} "
            + f"{fcolors[(n % len(fcolors))]}{s} {col}"
        )


def hbar_stack(data, columns=50, show_legend=True):
    """
    Print horizontal bar chart to stdout.

    Args:
        data (pd.DataFrame): Data to plot.
        columns (int): Maximum screen columns.

    Returns:
        None
    """
    labels = data.index
    norm_factor = data.T.sum().max()
    data_norm = data.apply(_normalize_data, args=([columns, norm_factor]))
    legend = {}
    print()
    for i in range(len(data_norm.index)):
        print(f"{co.Fore.BLUE}{str(labels[i])}: ", end="")
        sorted_row = data_norm.iloc[i].sort_values(ascending=False)
        for n, val in enumerate(sorted_row):
            fc = fcolors[n % len(fcolors)]
            label = sorted_row.index[n]
            if label not in legend:
                legend[label] = fc
            s = plot_unit * val
            print(f"{legend[label]}{s}", end="")
        print(f" {utils.time_print(data.iloc[i].sum())}")

    # color legend
    # pad legend to same column as actual plotted data
    # eg: "mon: " --> legend_pad = 5, "01-08 Jun: " --> legend_pad = 11
    if show_legend:
        legend_pad = len(data_norm.index.values[0]) + 2
        pad = " " * legend_pad
        print()
        print(pad, end="")
        i = 1
        rows = int(columns / 10)
        for label, fc in legend.items():
            if i % rows != 0:
                print(f"{fc}{plot_unit*3} {label} ", end="  ")
            else:
                print(f"\n{pad}{fc}{plot_unit*3} {label} ", end="  ")
            i += 1
        print()



import logging
from calendar import monthrange
from datetime import datetime

import click
import parsedatetime as pdt
from tabulate import tabulate

from yatta import db as db
from yatta.config import Config
from yatta.plotting import _preproc_data, _weekday_to_label
from yatta.utils import time_print

logger = logging.getLogger(__name__)
cal = pdt.Calendar()


@click.command()
@click.option(
    "-d", "--day", "period", help="Print today's timesheet.", flag_value="day",
)
@click.option(
    "-w",
    "--week",
    "period",
    help="Print this week's timesheet.",
    flag_value="week",
    default=True,
)
@click.option(
    "-s", "--start-date", help="Report data from this date onward.", default="now"
)
def timesheet(period, start_date):
    """
    View daily and weekly timesheet summary.
    """
    start_date = datetime(*cal.parse(start_date)[0][:6])
    day_of_week = start_date.weekday()
    year, month, day = start_date.timetuple()[:3]
    if period == "day":
        start_date = datetime(year, month, day)
    elif period == "week":
        # week straddles two months
        if day_of_week > day:
            diff = day_of_week - day
            month = month - 1
            # week straddles two years
            if month == 0:
                month = 12
                year = year - 1
            day = monthrange(year, month)[1] - diff
            start_date = datetime(year, month, day)
        else:
            start_date = datetime(year, month, day - day_of_week)
    elif period == "month":
        start_date = datetime(year, month, 1)
    else:
        logger.error("Invalid timesheet period!")

    query = db.get_records()
    data = _preproc_data(db.query_to_df(query), period, start_date)
    if not data.empty:
        if period == "day":
            data.index = [_weekday_to_label(day_of_week)]
            data["TOTAL"] = data.sum(axis=1)
            for col in data:
                data[col] = data[col].apply(time_print)
            print(
                "\n"
                + tabulate(
                    data,
                    headers=data.columns,
                    tablefmt=Config().get_user_value("formatting", "table_style"),
                )
            )
        elif period == "week":
            data["TOTAL"] = data.sum(axis=1)
            data.loc["TOTAL"] = data.sum()
            for col in data:
                data[col] = data[col].apply(time_print)
            print(
                "\n"
                + tabulate(
                    data,
                    headers=data.columns,
                    tablefmt=Config().get_user_value("formatting", "table_style"),
                )
            )
        else:
            logger.error("Invalid timesheet period!")
    else:
        print("No data for this period!")



import logging
import os

import pandas as pd
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
    create_engine,
    event,
    func,
)
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from tabulate import tabulate

from yatta import utils as utils
from yatta.config import Config
from yatta.utils import get_app_dirs


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


DATA_DIR, CONFIG_DIR, CACHE_DIR = get_app_dirs()
DB_PATH = os.path.join(DATA_DIR, "yatta.db")

logger = logging.getLogger(__name__)

engine = create_engine(f"sqlite:///{DB_PATH}", echo=False)
Sessionmkr = sessionmaker(bind=engine)
session = Sessionmkr()
Base = declarative_base()


# define tables / db objects
class Task(Base):
    __tablename__ = "tasks"
    # __table_args__ = {'extend_existing': True}
    name = Column(String, unique=True)
    id = Column(Integer, primary_key=True, autoincrement=True)
    tags = Column(String)
    description = Column(String)
    total = Column(Integer)

    records = relationship(
        "Record",
        primaryjoin="and_(Task.id == Record.task_id, Task.name == Record.task_name)",
        lazy="dynamic",
        cascade="all, delete-orphan",
    )

    def __repr__(self):
        return (
            f"<Task(name={self.name}, tags={self.tags}, "
            + f"description={self.description}, total={utils.time_print(self.total)})>"
        )

    def __str__(self):
        headers = ["ID", "Task", "Tags", "Description", "Total"]
        data = [
            [
                self.id,
                self.name,
                self.tags,
                self.description,
                utils.time_print(self.total),
            ],
        ]
        return "\n" + tabulate(
            data,
            headers=headers,
            tablefmt=Config().get_user_value("formatting", "table_style"),
        )


class Record(Base):
    __tablename__ = "records"
    id = Column(Integer, primary_key=True, autoincrement=True)
    task_id = Column(
        String, ForeignKey("tasks.id", onupdate="cascade", ondelete="cascade")
    )
    task_name = Column(
        String, ForeignKey("tasks.name", onupdate="cascade", ondelete="cascade")
    )
    start = Column(DateTime, nullable=False)
    end = Column(DateTime, nullable=False)
    duration = Column(Integer, nullable=False)

    def __repr__(self):
        return (
            f"<Record(task_name={self.task_name}, start={self.start} "
            + f"end={self.start}, duration={utils.time_print(self.duration)})>"
        )

    def __str__(self):
        h = ["record_id", "task_id", "task", "start", "end", "duration"]
        s = [
            [
                self.id,
                self.task_id,
                self.task_name,
                self.start,
                self.end,
                utils.time_print(self.duration),
            ],
        ]
        return "\n" + tabulate(
            s, headers=h, tablefmt=Config().get_user_value("formatting", "table_style"),
        )


# add the above mappings to database, if they don't already exist
Base.metadata.create_all(engine)


def add_record(task, record):
    session.add(task)
    session.commit()
    task.records.append(record)
    session.add(record)
    session.commit()


def get_records(record_id=None, task_name_or_id=None):
    if record_id:
        try:
            record_id = int(record_id)
            query = session.query(Record).filter(Record.id == record_id)
        except ValueError:
            logger.warning("Record ID must be an integer!")
            # query all records instead
            query = session.query(Record)
    elif task_name_or_id:
        # check if it's a name or an id
        try:
            task_name_or_id = int(task_name_or_id)
            query = session.query(Record).filter(Record.task_id == task_name_or_id)
        except ValueError:
            query = session.query(Record).filter(Record.task_name == task_name_or_id)
    else:
        query = session.query(Record)  # return all records
    return query


def get_tasks(task_name_or_id=None):
    if task_name_or_id:
        # check if it's a name or an id
        try:
            assert int(task_name_or_id)
            query = session.query(Task).filter(Task.id == task_name_or_id)
        except ValueError:
            query = session.query(Task).filter(Task.name == task_name_or_id)
    else:
        query = session.query(Task)  # return all tasks
    return query


def query_to_df(query):
    df = pd.read_sql(query.statement, query.session.bind, index_col="id")
    return df


def validate_task(task):
    """
    Make sure task object meets certain requirements.
    """
    # Ensure that task name cannot be converted to int (conflicts w/ task id)
    try:
        assert int(task.name)
        logger.warning("Cannot use integers as task names!")
        return False
    except ValueError:
        return True


def delete_task(task):
    session.delete(task)
    session.commit()


def delete_record(record):
    session.delete(record)
    session.commit()


def update_task_total(task):
    total = (
        session.query(func.sum(Record.duration))
        .filter(Record.task_id == task.id)
        .first()[0]
    )
    task.total = total
    session.commit()



import logging
import os
import time
from datetime import datetime

from appdirs import user_cache_dir, user_config_dir, user_data_dir

APP_NAME = "yatta"
CACHE_DIR = user_cache_dir(APP_NAME)
TMP_FILE = os.path.join(CACHE_DIR, "active_task")
PID_FILE = os.path.join(CACHE_DIR, "yatta.pid")

logger = logging.getLogger(__name__)


def get_app_dirs():
    """
    Make sure common app directories exists, create them if not
    """
    APP_NAME = "yatta"
    DATA_DIR = user_data_dir(APP_NAME)
    CONFIG_DIR = user_config_dir(APP_NAME)
    CACHE_DIR = user_cache_dir(APP_NAME)
    if not os.path.isdir(DATA_DIR):
        try:
            os.mkdir(DATA_DIR)
            logger.debug(f"Created data directory: {DATA_DIR}")
        except OSError:
            logger.error(f"Failed to create directory: {DATA_DIR}")
    if not os.path.isdir(CONFIG_DIR):
        try:
            os.mkdir(CONFIG_DIR)
            logger.debug(f"Created config directory: {CONFIG_DIR}")
        except OSError:
            logger.error(f"Failed to create directory: {CONFIG_DIR}")
    if not os.path.isdir(CACHE_DIR):
        try:
            os.mkdir(CACHE_DIR)
            logger.debug(f"Created cache directory: {CACHE_DIR}")
        except OSError:
            logger.error(f"Failed to create directory: {CACHE_DIR}")

    return (DATA_DIR, CONFIG_DIR, CACHE_DIR)


def time_div(count):
    min, sec = divmod(count, 60)
    hour, min = divmod(min, 60)
    return (hour, min, sec)


def time_format(hour, min, sec):
    return f"{hour:02}:{min:02}:{sec:02}"


def time_print(count):
    hour, min, sec = time_div(count)
    return time_format(hour, min, sec)


def time_figlet_print(font, count):
    """
    Print time as ASCII art.

    Args:
        font (pyfiglet.Figlet): A Figlet "font" object.
        count (int): Number of seconds.

    Returns:
        Figlet string
    """
    hour, min, sec = time_div(count)
    return font.renderText(time_format(hour, min, sec))


def stopwatch(taskname):
    start_time = datetime.now()
    while True:
        end_time = datetime.now()
        duration = (end_time - start_time).seconds
        _write_tmp_info(taskname, start_time, end_time, duration)
        time.sleep(1)


def _write_tmp_info(taskname, start, end, duration):
    with open(TMP_FILE, "w") as f:
        f.write(f"{start}\n{end}\n{duration}\n{taskname}\n{time_print(duration)}")


def _read_tmp_info():
    with open(TMP_FILE, "r") as f:
        # ignoring the formatting duration, this is intended for external
        # tools accessing the tmp file
        start, end, duration, taskname, _ = f.read().split("\n")
        start = datetime.strptime(start, "%Y-%m-%d %H:%M:%S.%f")
        end = datetime.strptime(end, "%Y-%m-%d %H:%M:%S.%f")
        duration = int(duration)
    return (start, end, duration, taskname)


def _last_day_of_month(any_day):
    next_month = any_day.replace(28) + datetime.timedelta(days=4)
    return next_month - datetime.timedelta(days=next_month.day)



import logging
from calendar import monthrange
from datetime import datetime

import click
import parsedatetime as pdt

from yatta import db as db
from yatta import plotting as plt
from yatta.config import Config

logger = logging.getLogger(__name__)

cal = pdt.Calendar()


@click.command()
@click.option("-d", "--day", "period", help="Plot today's timesheet.", flag_value="day")
@click.option(
    "-w",
    "--week",
    "period",
    help="Plot this week's timesheet.",
    flag_value="week",
    default=True,
)
@click.option(
    "-m", "--month", "period", help="Plot this month's timesheet.", flag_value="month"
)
@click.option(
    "-s", "--start-date", help="Plot data from this date onward.", default="now"
)
@click.option(
    "-c",
    "--columns",
    type=click.INT,
    help="Maximum columns on screen for plot to occupy.",
    default=Config().get_user_value("plotting", "columns"),
)
@click.option("--show-legend", is_flag=True, default=True)
def plot(period, start_date, columns, show_legend):
    """
    Visualize summary data.
    """
    # check if user wants to always show legend, change show_legend accordingly
    if not Config().user["plotting"]["show_legend"]:
        show_legend = False
    start_date = datetime(*cal.parse(start_date)[0][:6])
    day_of_week = start_date.weekday()
    year, month, day = start_date.timetuple()[:3]
    if period == "day":
        start_date = datetime(year, month, day)
    elif period == "week":
        # week straddles two months
        if day_of_week > day:
            diff = day_of_week - day
            month = month - 1
            # week straddles two years
            if month == 0:
                month = 12
                year = year - 1
            day = monthrange(year, month)[1] - diff
            start_date = datetime(year, month, day)
        else:
            start_date = datetime(year, month, day - day_of_week)
    elif period == "month":
        start_date = datetime(year, month, 1)
    else:
        logger.error("Invalid plot period!")

    query = db.get_records()
    data = plt._preproc_data(db.query_to_df(query), period, start_date)
    if not data.empty:
        if period == "day":
            plt.hbar(data, columns)
        elif period == "week":
            plt.hbar_stack(data, columns, show_legend)
        elif period == "month":
            plt.hbar_stack(data, columns, show_legend)
        else:
            logger.error("Invalid plot period!")
    else:
        print("No data for this period!")



import logging
import os
import time
from multiprocessing import Process

import click
from pyfiglet import Figlet

from yatta import db as db
from yatta.completion_helpers import get_matching_tasks
from yatta.config import Config
from yatta.daemon import daemon_start, dummy_stopwatch
from yatta.utils import get_app_dirs

DATA_DIR, CONFIG_DIR, CACHE_DIR = get_app_dirs()
PID_FILE = os.path.join(CACHE_DIR, "yatta.pid")

logger = logging.getLogger(__name__)


@click.command()
@click.argument("task", type=click.STRING, autocompletion=get_matching_tasks)
@click.option("-t", "--tags", default="", help="Add relevant tags to task.")
@click.option("-d", "--description", default="", help="Additional task info.")
@click.option(
    "-f", "--font", default="doom", help="Select figlet font (http://www.figlet.org/)."
)
@click.option("-b", "--background", help="Run yatta in the background.", is_flag=True)
def start(task, tags, description, font, background, **kwargs):
    """
    Start tracking a task.
    """
    # first check if there is already a task being recorded
    if not os.path.exists(PID_FILE):
        # create task if it doesn't exist
        task_name_or_id = task
        query = db.get_tasks(task_name_or_id)
        if not query.first():
            task = db.Task(name=task_name_or_id, tags=tags, description=description)
            task.total = 0
            db.validate_task(task)
            db.session.add(task)
            db.session.commit()
        else:
            task = query.first()
        font = Figlet(font=Config().get_user_value("formatting", "figlet_font"))
        if (
            background
            or Config().get_user_value("general", "run_in_background") == "true"
        ):
            daemon_start(task.name)
        else:
            Process(target=daemon_start, args=(task.name,)).start()
            # stagger processes so pidfile gets created first
            time.sleep(0.01)
            Process(target=dummy_stopwatch, args=(task.name, font)).start()
    else:
        print("You are already tracking a task. Try `yatta status` for more details.")



#!/usr/bin/env python
import logging

import click

from yatta import utils as utils
from yatta.commands import config, delete, edit, list, plot, start, timesheet
from yatta.daemon import daemon_status, daemon_stop

try:
    from importlib import metadata
except ImportError:  # for Python < 3.8
    import importlib_metadata as metadata

APP_NAME = "yatta"
DATA_DIR, CONFIG_DIR, CACHE_DIR = utils.get_app_dirs()

CLICK_CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"],)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
@click.version_option(version=metadata.version(APP_NAME))
@click.option("-l", "--log_level", default="warning")
def main(log_level):
    # setup logging
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {log_level}")
    logging.basicConfig(
        format="--> %(levelname)s in %(name)s: %(message)s", level=numeric_level
    )


@click.command()
def status():
    """
    Check status of active task.
    """
    daemon_status()


@click.command()
def stop():
    """
    Stop tracking the active task.
    """
    daemon_stop()


# add cli commands and command groups to main
main.add_command(start.start)
main.add_command(list.list)
main.add_command(edit.edit)
main.add_command(delete.delete)
main.add_command(plot.plot)
main.add_command(timesheet.timesheet)
main.add_command(config.config)
main.add_command(status)
main.add_command(stop)


if __name__ == "__main__":
    main()



import click
from tabulate import tabulate

from yatta import db as db
from yatta import utils as utils
from yatta.completion_helpers import get_matching_tasks
from yatta.config import Config


@click.group()
def list():
    """
    List information about tasks and records.
    """
    pass


@list.command()
@click.argument(
    "task_name", required=False, type=click.STRING, autocompletion=get_matching_tasks
)
def tasks(task_name=None):
    """
    List all tasks or show info for a particular task.
    """
    query = db.get_tasks(task_name)
    _tasks = db.query_to_df(query)
    _tasks["total"] = _tasks["total"].apply(utils.time_print)
    print(
        "\n"
        + tabulate(
            _tasks,
            headers=_tasks.columns,
            tablefmt=Config().get_user_value("formatting", "table_style"),
        )
    )


@list.command()
@click.argument("record_id", required=False)
@click.option(
    "-t",
    "--task",
    help="Task name or ID to get records from.",
    type=click.STRING,
    autocompletion=get_matching_tasks,
)
@click.option(
    "-m",
    "--max-entries",
    default=10,
    help="Number of recent records to return. (Default: 10)",
)
@click.option("-a", "--all", is_flag=True, help="List all records.")
def records(all, max_entries, record_id=None, task=None):
    """
    List recent records from all tasks or a particular task.
    """
    query = db.get_records(record_id=record_id, task_name_or_id=task)
    _records = db.query_to_df(query)
    if not all:
        _records = _records.tail(max_entries)
    _records["duration"] = _records["duration"].apply(utils.time_print)
    print(
        "\n"
        + tabulate(
            _records,
            headers=_records.columns,
            tablefmt=Config().get_user_value("formatting", "table_style"),
        )
    )



"""
Functions to generate custom tab completions for click.
"""
import tabulate
from pyfiglet import Figlet

from yatta import db as db
from yatta.config import Config


def get_matching_tasks(ctx, args, incomplete):
    task_names = [task.name for task in db.get_tasks().all()]
    return [task_name for task_name in task_names if incomplete in task_name]


def get_figlet_fonts(ctx, args, incomplete):
    f = Figlet(Config().get_user_value("formatting", "figlet_font"))
    fonts = sorted(f.getFonts())
    return [font for font in fonts if incomplete in font]


def get_table_formats(ctx, args, incomplete):
    return [fmt for fmt in tabulate.tabulate_formats if incomplete in fmt]



import logging
import os

from tomlkit import document, dumps, parse, table

from yatta.utils import get_app_dirs

DATA_DIR, CONFIG_DIR, CACHE_DIR = get_app_dirs()
CONFIG_FILE = os.path.join(CONFIG_DIR, "settings.toml")

logger = logging.getLogger(__name__)


class Config(object):
    """
    A class to handle app configuration.
    """

    default_dict = {
        "general": {"run_in_background": "false"},
        "formatting": {"figlet_font": "doom", "table_style": "pretty"},
        "plotting": {"columns": 75, "show_legend": "true"},
    }

    def __init__(self):
        self.default = construct_toml(self.default_dict)
        self.user = self._generate_user_config()

    def __str__(self):
        return dumps(self.user)

    def _generate_user_config(self):
        if os.path.exists(CONFIG_FILE):
            # read user settings from file
            try:
                with open(CONFIG_FILE, "r") as f:
                    content = f.read()
                    config = parse(content)
            except Exception as e:
                logger.error(e)
            # check for new default settings
            for category, settings in self.default_dict.items():
                if category not in config.keys():
                    config.add(category, table())
                else:
                    for setting, value in settings.items():
                        if setting not in config[category].keys():
                            config[category][setting] = value
            # remove deprecated settings
            for category, settings in config.items():
                # remove deprecated tables
                if category not in self.default_dict.keys():
                    config.remove(category)
                # remove deprecated settings within tables
                else:
                    for setting in settings.keys():
                        if setting not in self.default_dict[category]:
                            config[category].remove(setting)
        else:
            # construct the default config toml if no config file exists
            config = construct_toml(self.default_dict)
        # write to file
        self._write_to_config(config)
        return config

    def _write_to_config(self, toml_doc):
        with open(CONFIG_FILE, "w") as f:
            f.write(dumps(toml_doc))

    def _check_new_settings(self):
        """
        When new settings are added to `default_dict` in later versions, need to be
        able to add these into the user config file without wiping the file.
        """
        for category, settings in self.default_dict.items():
            if category not in self.user.keys():
                self.user.add(category, table())
                for setting, value in settings.items():
                    self.user[category][setting] = value
        self._write_to_config(self.user)

    def _remove_old_settings(self):
        pass

    def restore_defaults(self):
        self.user = construct_toml(self.default_dict)
        self._write_to_config(self.user)

    def get_user_value(self, table, key):
        return self.user[table][key]

    def set_user_value(self, table, key, value):
        self.user[table][key] = value
        self._write_to_config(self.user)


def construct_toml(config_dict):
    toml_doc = document()
    tables = {}
    for category in config_dict.keys():
        tables[category] = table()
        for setting, value in config_dict[category].items():
            tables[category].add(setting, value)
        toml_doc.add(category, tables[category])
    return toml_doc



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['yatta', 'yatta.commands']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.3,<0.5.0',
 'importlib-metadata>=1.6.0,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'parsedatetime>=2.5,<3.0',
 'pyfiglet>=0.8.post1,<0.9',
 'sqlalchemy>=1.3.17,<2.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tomlkit>=0.6.0,<0.7.0']

entry_points = \
{'console_scripts': ['yatta = yatta.console:main']}

setup_kwargs = {
    'name': 'yatta',
    'version': '0.2.4',
    'description': 'Yet Another Time Tracking Application',
    'long_description': None,
    'author': 'Rusty Roberts',
    'author_email': 'rust.roberts@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rhroberts/yatta',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import logging

import click

from yatta import db as db
from yatta.completion_helpers import get_matching_tasks

logger = logging.getLogger(__name__)


@click.group()
def delete():
    """
    Delete tasks or records.
    """


@delete.command()
@click.argument(
    "task_name_or_id", nargs=-1, type=click.STRING, autocompletion=get_matching_tasks
)
@click.confirmation_option(prompt="Are you sure you want to delete specified task(s)?")
def tasks(task_name_or_id):
    """
    Delete a task and all its associated records.
    """
    for name_id in task_name_or_id:
        query = db.get_tasks(name_id)
        _task = query.first()
        if not _task:
            print(f"Task '{name_id}' does not exist.")
            return
        try:
            # delete task from tasks table; cascade delete associated records
            db.delete_task(_task)
            print(f"Deleted task {name_id}!")
        except Exception as e:
            db.session.rollback()
            logger.error(e)


@delete.command()
@click.argument("record_id", nargs=-1)
@click.confirmation_option(prompt="Are you sure you want to delete specifed record(s)?")
def records(record_id):
    """
    Delete a single record by its ID.
    """
    for _id in record_id:
        query = db.get_records(record_id=_id)
        _record = query.first()
        if not _record:
            print(f"Record id '{_id}' does not exist.")
        try:
            db.delete_record(_record)
            print(f"Deleted record {_id}!")
        except Exception as e:
            db.session.rollback()
            logger.error(e)
